# Copyright (C) 2005-2006 by Dr. Dieter Maurer, Eichendorffstr. 23, D-66386 St. Ingbert, Germany
# see "LICENSE.txt" for details
#       $Id: testISearch.py,v 1.1.1.1 2008/06/27 19:21:39 dieter Exp $
'''IncrementalSearch tests.'''

from unittest import TestCase, TestSuite, makeSuite, TextTestRunner
from BTrees.IIBTree import IISet
from BTrees.LLBTree import LLSet
from BTrees.OOBTree import OOSet

from dm.incrementalsearch import ISearch, IBTree, Enumerator, \
     INLIST_SUCCESS, CANDIDATE, NOT_INLIST, AT_END, \
     IOr, INot, IAnd, \
     IOr_int, IAnd_int, \
     INT, OBJECT, LONG, \
     Unspecified

class _ObjTests:
  KEYTYPE = OBJECT

class _LongTests:
  KEYTYPE = LONG

class IOrTests(TestCase):
  KEYTYPE = INT

  def setUp(self):
    if self.KEYTYPE == INT: set = IISet
    elif self.KEYTYPE == LONG: set = LLSet
    else: set = OOSet
    self._search = IOr(
      self.KEYTYPE,
      IBTree(set(range(5,20,5)), self.KEYTYPE),
      IBTree(set(range(2,20,2)), self.KEYTYPE),
      )
    self._search.complete()

  def testResult(self):
    self.assertEqual(self._search.asSet().keys(),
                     IISet(range(5,20,5) + range(2,20,2)).keys()
                     )
   
  def testToTraversal(self):
    search = self._search
    l = []; i = 0
    while True:
      cr = search.advanceTo(i)
      self.assertEqual(i >= 19, cr == AT_END)
      if cr == AT_END: break
      if cr == INLIST_SUCCESS: l.append(search.value)
      if cr == CANDIDATE: i = search.value
      else: i += 1
    self.assertEqual(l, IISet(range(5,20,5) + range(2,20,2)).keys())

  def testFromJumping(self):
    search = self._search
    for s,r in ((4,5), (6,8), (14,15), (15,16)):
      self.assertEqual(search.advanceFrom(s, Unspecified), INLIST_SUCCESS)
      self.assertEqual(search.value, r)
    self.assertEqual(search.advanceFrom(18, Unspecified), AT_END)

  def testFromLimit(self):
    search = self._search
    self.assertEqual(search.advanceFrom(2,4), CANDIDATE)
    self.assertEqual(search.value, 4)
    self.assertEqual(search.advanceFrom(8,11), INLIST_SUCCESS)
    self.assertEqual(search.value, 10)
    self.assertEqual(search.advanceFrom(14,16), INLIST_SUCCESS)
    self.assertEqual(search.value, 15)

  def testFromToInterleaving(self):
    search = self._search
    self.assertEqual(search.advanceFrom(8,10), CANDIDATE)
    self.assertEqual(search.value, 10)
    self.assertEqual(search.advanceTo(10), INLIST_SUCCESS)
    self.assertEqual(search.advanceFrom(10,11), CANDIDATE)
    self.assertEqual(search.value, 12)
    self.assertEqual(search.advanceTo(12), INLIST_SUCCESS)
    self.assertEqual(search.advanceFrom(12,13), CANDIDATE)
    self.assertEqual(search.value, 14)
    self.assertEqual(search.advanceTo(15), INLIST_SUCCESS)

class IOrTests_obj(_ObjTests, IOrTests): pass
class IOrTests_long(_LongTests, IOrTests): pass

class INotTests(TestCase):
  KEYTYPE = INT

  def setUp(self):
    if self.KEYTYPE == INT: set = IISet
    elif self.KEYTYPE == LONG: set = LLSet
    else: set = OOSet
    self._search = INot(
      IBTree(set(range(4, 12, 4)), self.KEYTYPE),
      Enumerator(set(range(2, 12, 2))),
      )

  def testResult(self):
    self.assertEqual(self._search.asSet().keys(), [2, 6, 10])

  def testToTraversal(self):
    search = self._search
    for i in range(2, 12, 2):
      cr = search.advanceTo(i)
      self.assertEqual(i > 10, cr == AT_END)
      self.assertEqual(i in [2, 6, 10], cr == INLIST_SUCCESS)
      self.assertEqual(i <= 10 and i not in [2, 6, 10], cr == NOT_INLIST)

  def testFromLimit(self):
    search = self._search
    self.assertEqual(search.advanceFrom(2,6), CANDIDATE)
    self.assertEqual(search.value, 6)
    self.assertEqual(search.advanceTo(6), INLIST_SUCCESS)
    self.assertEqual(search.value, 6)
    self.assertEqual(search.advanceFrom(7,11), INLIST_SUCCESS)
    self.assertEqual(search.value, 10)
    self.assertEqual(search.advanceFrom(11,13), AT_END)

class INotTests_obj(_ObjTests, INotTests): pass
class INotTests_long(_LongTests, INotTests): pass

class IFilterTests(TestCase):
  KEYTYPE = INT

  def setUp(self):
    if self.KEYTYPE == INT: set = IISet
    elif self.KEYTYPE == LONG: set = LLSet
    else: set = OOSet
    self._search = IFilter(
      self.KEYTYPE,
      lambda i, inot=range(4, 12, 4): i not in inot,
      Enumerator(set(range(2, 12, 2))),
      )

class IFilterTests_obj(_ObjTests, IFilterTests): pass
class IFilterTests_long(_LongTests, IFilterTests): pass



class IntegrationTests(TestCase):
  KEYTYPE = INT

  def testAndOrNot(self):
    if self.KEYTYPE == INT: set = IISet
    elif self.KEYTYPE == LONG: set = LLSet
    else: set = OOSet
    enumerator = Enumerator(set(range(0,17)))
    search = IOr(self.KEYTYPE,
      IBTree(set(range(5,17,5)), self.KEYTYPE),
      INot(IBTree(set(range(5,11)), self.KEYTYPE), enumerator),
      )
    search.complete()
    search = IAnd(self.KEYTYPE,
      search,
      INot(IBTree(set(range(1,17,2)), self.KEYTYPE), enumerator),
      )
    search.complete()
    self.assertEqual(search.asSet().keys(), [0,2,4,10,12,14,16])

class IntegrationTests_obj(_ObjTests, IntegrationTests): pass
class IntegrationTests_long(_LongTests, IntegrationTests): pass


class _HitNohitISearch(ISearch):
  '''auxiliary class for covering IOr test.

  Does not report candidates but only hits and nohits.
  '''
  def __init__(self, keytype, set):
    ISearch.__init__(self, keytype)
    self._set = set
    self.estimatedSize = len(set)

  def advanceTo(self, end):
    return self._locate(end)

  def advanceFrom(self, start, upto):
    return self._locate(start, False, upto)

  def _locate(self, key, include=True, limit=Unspecified):
    if key is Unspecified: key = None
    keys = list(self._set.keys(key))
    if not include and key in keys: keys.remove(key) # 21
    if not keys: return AT_END
    if include: cr = key != keys[0] and NOT_INLIST or INLIST_SUCCESS
    else: cr = limit is not Unspecified and keys[0] >= limit and CANDIDATE or INLIST_SUCCESS
    self.classification = cr
    if cr <= CANDIDATE: self.value = keys[0]
    else: self.value = key
    return cr

class IOrCoverTests(TestCase):
  KEYTYPE = INT

  def setUp(self):
    if self.KEYTYPE == INT: set = IISet
    elif self.KEYTYPE == LONG: set = LLSet
    else: set = OOSet
    s1 = set((0, 5, 10))
    s2 = set((0,3,6,10,12))
    self._search = IOr(self.KEYTYPE,
                       IBTree(s1, self.KEYTYPE),
                       _HitNohitISearch(self.KEYTYPE, s2),
                       )
    self._search.complete()
    r = set(s1)
    r.update(s2)
    self._result = r

  def testAdvanceTo(self):
    r = self._result; s = self._search; to = s.advanceTo
    for i in range(r.minKey(), r.maxKey()+1):
      cr = to(i)
      self.assertEqual(cr, i not in r and NOT_INLIST or INLIST_SUCCESS)
      if cr == INLIST_SUCCESS: self.assertEqual(s.value, i)
    self.assertEqual(to(r.maxKey()+1), AT_END)

  def testEnumerate(self):
    self.assertEqual(list(self._result), self._search.asSet().keys())

  def testAdvanceFrom(self):
    s = self._search; fr = s.advanceFrom
    def check(start, upto, result=None):
      cr = fr(start, upto)
      self.assert_((result is not None) == (cr == INLIST_SUCCESS))
      if cr == INLIST_SUCCESS: self.assertEqual(result, s.value)
    check(1,3)
    check(1,4,3)
    check(3,7,5)
    check(5,7,6)
    check(6,10)
    check(6,13, 10)
    check(11,13, 12)
    self.assertEqual(fr(12,14), AT_END)

class IOrCoverTests_obj(_ObjTests, IOrCoverTests): pass
class IOrCoverTests_long(_LongTests, IOrCoverTests): pass


class _ISearchCandidate1Hit2(ISearch):
  def __init__(self):
    ISearch.__init__(self, INT)

  def advanceTo(self, end):
    self.value = end
    cl = end != 2 and NOT_INLIST or INLIST_SUCCESS
    self.classification = cl
    #print 'advanceTo', end, cl
    return cl

  def advanceFrom(self, fr, upto):
    if fr is Unspecified or fr < 1:
      if upto is Unspecified or upto > 2: v = 2; cl = INLIST_SUCCESS
      elif upto <= 1: v = 1; cl = CANDIDATE
      else: v = 2; cl = CANDIDATE
    elif fr < 2:
      if upto is Unspecified or upto > 2: v = 2; cl = INLIST_SUCCESS
      else: v = upto; cl = CANDIDATE
    else: cl = AT_END; v = 2
    self.classification = cl
    self.value = v
    #print 'advanceFron', fr, upto, cl ,v
    return cl

class IOrToCandidateBug(TestCase):
  def testToCandidateBug(self):
    s = IISet((0,))
    ss = IOr_int(IBTree(s), _ISearchCandidate1Hit2())
    ss.complete()
    self.assertEqual(ss.advanceFrom(Unspecified, Unspecified), INLIST_SUCCESS)
    self.assertEqual(ss.value, 0)
    self.assertEqual(ss.advanceFrom(ss.value, Unspecified), INLIST_SUCCESS)
    self.assertEqual(ss.value, 2)
    self.assertEqual(ss.advanceFrom(ss.value, Unspecified), AT_END)



class TestCompletion(TestCase):
  set = [5, [1,2,3], 4, [-1, 6], -1]

  def testAnd(self):
    self._check('and')

  def testOr(self):
    self._check('or')

  def _check(self, op):
    search, flat = self._build(op, self.set)
    sizes = [s.estimatedSize for s in search.unmantle() if s >= 0]
    proper = [i for i in flat if i >= 0]
    no_improper = len(flat) - len(proper)
    if no_improper:
      self.assertEqual(sizes[-no_improper:], [-1] * no_improper)
      sizes = sizes[:-no_improper]
    proper.sort()
    if op == 'or': proper.reverse()
    self.assertEqual(proper, sizes)

  def _build(self, op, set):
    class _Proxy(ISearch):
      def __init__(self, size):
        self._s = size
        ISearch.__init__(self, INT)
        self.estimatedSize = size
    sl = []; il = []
    for s in set:
      if isinstance(s, list):
        search, sizes = self._build(op, s)
        sl.append(search); il += sizes
      else: sl.append(_Proxy(s)), il.append(s)
    cl = op == 'and' and IAnd_int or IOr_int
    search = cl(*sl); search.complete()
    return search, il

    

def test_suite():
  md = globals()
  cls = []
  for cl in [
    IOrTests,
    IOrCoverTests,
    INotTests,
    IFilterTests,
    IntegrationTests,
    ]:
    cls.append(cl)
    cls.append(md[cl.__name__ + '_obj'])
    cls.append(md[cl.__name__ + '_long'])
  cls.append(TestCompletion)
  cls.append(IOrToCandidateBug)
  return TestSuite([makeSuite(cl) for cl in cls])

def main():
  tester = TextTestRunner()
  tester.run(test_suite())

if __name__ == '__main__': main()




# Copyright (C) 2005 by Dr. Dieter Maurer, Eichendorffstr. 23, D-66386 St. Ingbert, Germany
# see "LICENSE.txt" for details
#       $Id: testIBTrees.py,v 1.1.1.1 2008/06/27 19:21:39 dieter Exp $
'''Tests for BTrees based incremental searches.'''

from unittest import TestCase, TestSuite, makeSuite, TextTestRunner

import BTrees
from BTrees.IIBTree import IISet, IITreeSet
from BTrees.IOBTree import IOBucket, IOBTree
from BTrees.OOBTree import OOBucket, OOBTree, OOTreeSet, OOSet
from BTrees.LLBTree import LLSet, LLTreeSet
from BTrees.LOBTree import LOBucket, LOBTree

from dm.incrementalsearch import IBTree, Unspecified, \
     INLIST_SUCCESS, INLIST, AT_END, \
     INT, OBJECT, LONG, \
     intersection


class BTreeISearchTestBase(TestCase):
  '''test suite controlled by *keys* and *split*.'''

  # to be defined by deriving classes
  keys = None # keys the constructed BTree should have
  split = None # where the tree should have subtrees

  KEYTYPE = INT

  SET_CLASS = IISet


  def setUp(self):
    keys = self.keys
    set =  self._set = self.SET_CLASS(keys)
    tree, _ = self._makeTree(set, self.split)
    self._search = IBTree(tree, self.KEYTYPE)

  def _makeTree(self, set, split, next=None):
    if self.KEYTYPE == INT: Bucket = IOBucket; Tree = IOBTree
    elif self.KEYTYPE == LONG: Bucket = LOBucket; Tree = LOBTree
    else: Bucket = OOBucket; Tree = OOBTree 
    keys = list(set.keys())
    n = len(set)
    if split:
      i = 0; children = []
      for s in split:
        grandChildren = []
        while True:
          if i >= n: break
          x = set[i]
          if x >= s: break
          grandChildren.append(x); i += 1
        children.append(grandChildren)
        children.append(s)
      children.append(set.keys(split[-1]))
      for i in range(len(children)-1, -1, -2):
        tree = Bucket(); items = []
        for k in children[i]: items += [k, None]
        S = (tuple(items),)
        if next is not None: S += (next,)
        tree.__setstate__(S)
        next = children[i] = tree
      tree = Tree()
      tree.__setstate__((tuple(children), children[0]))
      return tree, children[0]
    else:
      tree = Bucket(); items = []
      for k in keys: items += [k, None]
      tree.__setstate__((tuple(items),))
      return tree, None

  def testTraverseFrom(self):
    val = Unspecified; search = self._search 
    for k in self.keys:
      cr = search.advanceFrom(val, Unspecified)
      self.assertEqual(cr, INLIST_SUCCESS)
      self.assertEqual(search.value, k)
      val = k
    self.assertEqual(search.advanceFrom(val, Unspecified), AT_END)

  def testTraverseTo(self):
    search = self._search 
    for k in self.keys:
      self.assertEqual(search.advanceTo(k), INLIST_SUCCESS)
      self.assertEqual(search.value, k)

  def testToWithSkipping(self):
    self._check(3, 4, 25, 30, 31)

  def testToBoundaries(self):
    self._check(14, 15, 28, 30, 31, 40, 41)

  def testFromSkipping(self):
    self._check((3,4), 4, (4,25), 5, (25,26), (29,Unspecified), (30,Unspecified))

  def testFromBoundaries(self):
    self._check((14,15), (15,16), (29,31), (29,36))

  def testFromBoundaries2(self):
    self._check((15,20))

  def _check(self, *cmds):
    current = Unspecified; cr = None; set = self._set; search = self._search
    for cmd in cmds:
      args = cmd
      if isinstance(cmd, tuple): val = cmd[0]; fun = search.advanceFrom
      else: val = cmd; fun = search.advanceTo; args = cmd,
      cr = fun(*args)
      hits = set.keys(val)
      if not hits: self.assertEqual(cr, AT_END); continue
      if isinstance(cmd, tuple):
        if hits[0] == val: hits = hits[1:]
        if not hits: self.assertEqual(cr, AT_END); continue
        else:
          current = hit = hits[0]; upto = cmd[1]
          self.assertEqual(search.value, hit)
          if upto is None or hit < upto: self.assertEqual(cr, INLIST_SUCCESS)
          else: self.assertEqual(cr, INLIST)
      else:
        current = hit = hits[0]
        self.assertEqual(search.value, hit)
        self.assertEqual(cr, hit > val and INLIST or INLIST_SUCCESS)

  def testIteration(self):
    self.assertEqual(list(self._set), list(self._search))


class _ObjTestBase:
  KEYTYPE = OBJECT

class _LongTestBase:
  KEYTYPE = LONG
  SET_CLASS = LLSet


class TestOneLevel(BTreeISearchTestBase):
  keys = range(1,31)

class TestOneLevel_obj(_ObjTestBase, TestOneLevel): pass
class TestOneLevel_long(_LongTestBase, TestOneLevel): pass

class TestSecondTrivialLevel(TestOneLevel):
  split = range(2,31)

class TestSecondTrivialLevel_obj(_ObjTestBase, TestSecondTrivialLevel): pass
class TestSecondTrivialLevel_long(_LongTestBase, TestSecondTrivialLevel): pass

class TestRegularSplit(TestOneLevel):
  slit = (4, 25,)

class TestRegularSplit_obj(_ObjTestBase, TestRegularSplit): pass
class TestRegularSplit_long(_LongTestBase, TestRegularSplit): pass

class TestIrregular(BTreeISearchTestBase):
  keys = range(2,15,2) + range(15,30,2) + [31, 35, 40]
  split = (15, 30,)

class TestIrregular_obj(_ObjTestBase, TestIrregular): pass
class TestIrregular_long(_LongTestBase, TestIrregular): pass

class TestLongLargeValues(BTreeISearchTestBase):
  keys = range(1,31) + range(0x10000000, 0x10000010)
  SET_KLASS = LLSet
  


class TestTrueTree(BTreeISearchTestBase):
  # to be overridden
  keys = None
  class_ = IOBTree
  mapping = True

  def setUp(self):
    keys = self.keys
    self._set = IISet(keys)
    if self.mapping: items = map(None, keys, ())
    else: items = keys
    self._search = IBTree(self.class_(items), self.KEYTYPE)


class TestMedBTree(TestTrueTree):
  keys = range(1, 70)

class TestMedBTree_obj(TestMedBTree):
  class_ = OOBTree
  KEYTYPE = OBJECT


class _DerivedBTree(IOBTree):
  '''derived from IOBTree.'''

class TestMedDerivedBTree(TestMedBTree):
  class_ = _DerivedBTree


class TestMedBTree2(TestTrueTree):
  keys = range(1, 150, 2)

class TestMedBTree2_obj(TestMedBTree2):
  class_ = OOBTree
  KEYTYPE = OBJECT

class TestSmallBTree(TestTrueTree):
  keys = range(1,10)

class TestSmallBTree_obj(TestSmallBTree):
  class_ = OOBTree
  KEYTYPE = OBJECT


class TestBucket(TestTrueTree):
  keys = range(1, 70)
  class_ = IOBucket

class TestBucket_obj(TestBucket):
  class_ = OOBucket
  KEYTYPE = OBJECT
 
class TestTreeSet(TestTrueTree):
  keys = range(1, 500)
  class_ = IITreeSet
  mapping = False

class TestTreeSet_obj(TestTreeSet):
  class_ = OOTreeSet
  KEYTYPE = OBJECT

class TestSet(TestTreeSet):
  class_ = IISet

class TestSet_obj(TestTreeSet):
  class_ = OOSet
  KEYTYPE = OBJECT

class TestEmptyTree(TestTrueTree):
  keys = ()

  def testEstimatedSize(self):
    self.assertEqual(self._search.estimatedSize, 0)


class TestDepthThree(BTreeISearchTestBase):
  keys1 = range(2,15,2); split1 = [10]
  keys2 = range(15,30,2); split2 = [19]
  keys3 = [31, 35, 40]; split3 = [38]
  keys = keys1 + keys2 + keys3
  split = [15, 30]

  set_class_ = IISet
  tree_class_ = IOBTree

  def setUp(self):
    self._set = self.set_class_(self.keys1 + self.keys2 + self.keys3)
    mkt = self._makeTree
    tree3, leaf3 = mkt(self.set_class_(self.keys3), self.split3)
    tree2, leaf2 = mkt(self.set_class_(self.keys2), self.split2, leaf3)
    tree1, leaf1 = mkt(self.set_class_(self.keys1), self.split1, leaf2)
    tree = self.tree_class_(); split = self.split
    tree.__setstate__(((tree1, split[0], tree2, split[1], tree3), leaf1))
    self._search = IBTree(tree)

class TestDepthThree_obj(TestDepthThree):
  KEYTYPE = OBJECT
  set_class_ = OOSet
  tree_class_ = OOBTree

### Intersection tests
class IntersectionTestBase(TestCase):
  # to be overridden in deriving classes
  KEYTYPE = INT
  class_ = None
  mapping = None
  module = None

  set = range(1000)

  def testEmptyIntersection(self):
    self.assertRaises(ValueError, intersection, self.KEYTYPE)

  def testUnaryIntersection(self):
    self._check(1)

  def testEqualIntersection(self):
    self._check(1, 1)

  def testSubsetIntersection(self):
    self._check(2, 1)
    self._check(1, 2)

  def testEmptyResult(self):
    self._check(2, (2,1))

  def testTripleIntersection(self):
    self._check(2, 3, 5)

  def testSmallResult(self):
    s = [100, 200]
    self._check(s, 1)
    self._check(1, s)

  def _check(self, *args):
    args = map(self._makeSet, args)
    result = None; bti = self.module.intersection
    for set in args: result = bti(result, set)
    self.assert_(list(result.keys()) == intersection(self.KEYTYPE, *args).keys())

  def _makeSet(self, spec):
    '''make a set from *spec*.

    *spec* can be an integer (meaning: 'spec * self._set')
    a pair (*f*, *o*) (meaning 'f * self._set + o')
    or a list (meaning itself).
    '''
    set = self.set
    if isinstance(spec, int): spec = [spec * x for x in set]
    elif isinstance(spec, tuple):
      f, o = spec
      spec = [f * x + o for x in set]
    if self.mapping: spec = map(None, spec, ())
    return self.class_(spec)


class TestSetIntersection(IntersectionTestBase):
  class_ = IISet
  mapping = False
  module = BTrees.IIBTree

class TestSetIntersection_obj(TestSetIntersection):
  KEYTYPE = OBJECT
  class_ = OOSet
  module = BTrees.OOBTree

class TestSetIntersection_long(TestSetIntersection):
  KEYTYPE = LONG
  class_ = LLSet
  module = BTrees.LLBTree

class TestTreeSetIntersection(TestSetIntersection):
  class_ = IITreeSet

class TestTreeSetIntersection_obj(TestTreeSetIntersection):
  KEYTYPE = OBJECT
  class_ = OOTreeSet
  module = BTrees.OOBTree

class TestBucketIntersection(IntersectionTestBase):
  class_ = IOBucket
  module = BTrees.IOBTree
  mapping = True

class TestBucketIntersection_obj(TestBucketIntersection):
  KEYTYPE = OBJECT
  class_ = OOBucket
  module = BTrees.OOBTree

class TestBTreeIntersection(TestBucketIntersection):
  class_ = IOBTree

class TestBTreeIntersection_obj(TestBTreeIntersection):
  KEYTYPE = OBJECT
  class_ = OOBTree
  module = BTrees.OOBTree
    

def test_suite():
  md = globals()
  cls = []
  for cl in [
    TestOneLevel,
    TestSecondTrivialLevel,
    TestRegularSplit,
    TestIrregular,
    TestMedBTree,
    TestMedBTree2,
    TestSmallBTree,
    TestBucket,
    TestTreeSet,
    TestSet,
    TestDepthThree,
    TestSetIntersection,
    TestTreeSetIntersection,
    TestBucketIntersection,
    TestBTreeIntersection,
    ]:
    cls.append(cl)
    cls.append(md[cl.__name__ + '_obj'])
  for cl in [
    TestOneLevel,
    TestSecondTrivialLevel,
    TestRegularSplit,
    TestIrregular,
    TestSetIntersection,
    ]:
    cls.append(md[cl.__name__ + '_long'])
  cls.append(TestMedDerivedBTree)
  cls.append(TestEmptyTree)
  return TestSuite([makeSuite(cl) for cl in cls])

def main():
  tester = TextTestRunner()
  tester.run(test_suite())

if __name__ == '__main__': main()



# Copyright (C) 2005-2008 by Dr. Dieter Maurer, Illtalstr. 25, D-66571 Bubach, Germany
# see "LICENSE.txt" for details
#       $Id: __init__.py,v 1.1.1.1 2008/06/27 19:21:39 dieter Exp $
""""dm.incrementalsearch"

Incremental search is an efficient low level search execution engine
on top of "ZODB.BTrees". Its primary purpose is to incrementally
perform searches involving "and", "or" and "not" queries over
"ZODB.BTrees".

Incremental search is highly efficient for specific "and" queries
because it interleaves index lookup with the incremental determination
of hits. Often only small parts of the document lists involved
in the query need to be read. This can drastically reduce
the amount of loads and therefore IO which usually dominates
the total search time at least in distributed ZEO environments.
Furthermore, it makes more efficient use of the various ZODB
caches.


**ATTENTION** "incrementalsearch" is likely to fail when you
enumerate a "XXBTree" containing "None". This is caused by
a bug in the "XXBTree.keys(minKey, maxKey)" specification
which interprets a "None" value for "minKey" and "maxKey"
as "no restriction".


The central concept in the incremental search framework
is the "ISearch". An "ISearch" is similar to an iterator over
a sorted list of keys. Unlike an iterator, "ISearch"
has two methods to move forward over the list:
"advanceTo(key)" and
"advanceFrom(fromKey, limitKey)".
"advanceTo" tries to move (forward) in the sorted list to "key",
"advanceFrom" tries to move (forward) to the smallest key in the list
which is larger than "fromKey". The sorted list over which an
"ISearch" iterates is usually not yet fully materialized but
computed incrementally. Because the determination of the next
list element can be expensive, "limitKey" can be used to
indicate that only keys in the open interval between "fromKey"
and "limitKey" are of interest. There is a constant "Unspecified"
which can be passed as argument for "fromKey" or "limitKey".
When used for "fromKey", it is considered smaller than any key
in the list; when used for "limitKey", it is considered larger than
any key in the list.


"advanceTo" and "advanceFrom" are only allowed to move forward
in the following precise sense:

  1. when *key* is passed as argument to "advanceTo" (for an
     "ISearch" instance), then any value later passed
     as "key" to "advanceTo" or as "fromKey" to "advanceFrom"
     (for this same "ISearch" instance) must not be smaller
     than *key*

  2. when *fromKey* is passed as argument to "advanceFrom" (for an
     "ISearch" instance), then any value later passed
     as "key" to "advanceTo" or as "fromKey" to "advanceFrom"
     (for this same "ISearch" instance) must be (strictly) larger
     than *fromKey*.

A debug version may check this condition and raise an exception.
However, an "ISearch" is explicitely entitled to assume this
condition without check (and therefore to deliver wrong results
if the condition is not fullfilled).


An "ISearch" has two attributes "value" and "classification".
"value" indicates the instances current value, "classification"
provides information about whether "value" is in the list.
"classification" can have the following increasing integer constants as values:

  "INLIST_SUCCESS" (the integer 0) -- "value" is in the list (and the
    last call to an "advance*" method (for this instance)
    was "completely successful").

  "INLIST" -- "value" is in the list  (but  the
    last call to an "advance*" method (for this instance)
    was not "completely successful").

  "CANDIDATE" -- "value" may or may not be in the list.

  "NOT_INLIST" -- "value" is no longer valid
    Note: "advanceFrom" must not return this value

  "AT_START" -- iteration over the list has not yet started, "value" is undefined

  "AT_END" -- the list is exhausted, there are no more hits, "value" is undefined, no more "advance*" calls are allowed.

For convenience, the "advance*" methods return the classification value.


The "ISearch" attribute "keytype" can have the constants "INT"
and "OBJECT" as values. When "ISearch" instances are combined,
their "keyType" must be the same.

"ISearch" has the attribute "estimatedSize" containing an estimation
of the underlying lists size or "-1", if no reasonable estimation
can be obtained. It is used only for optimization heuristics.
"ISearch" may grow further attributes for optimization in the future.


The "ISearch" method "asSet" returns all (remaining) elements
in the underlying list as an "IISet" or "OOSet" (depending on
"keyType"). Note that for efficiency reasons, the "asSet" method
of "IBTree" returns the tree itself (which might not be a set).
"""

__all__ = (
  'ISearch IBTree '
  'IAnd IAnd_int IAnd_obj IAnd_long'
  'IOr IOr_int, IOr_obj IOr_long'
  'INot '
  'IFilter IFilter_int IFilter_obj IFilter_long'
  'INT OBJECT '
  'Unspecified '
  'INLIST_SUCCESS INLIST CANDIDATE NOT_INLIST AT_START AT_END '
  'Enumerator '
  'intersection intersection_int, intersection_obj intersection_long'
  ).split()

# the classification values -- must agree with 'enum Classification' in
# "ISearch.h"
INLIST_SUCCESS = 0
INLIST = 1
CANDIDATE = 2
NOT_INLIST = 3
AT_START = 4
AT_END = 5

# the keytype values -- must agree with 'enum Keytype' in "ISearch.h"
# UNDEFINED = 0
INT = 1
OBJECT = 2
LONG = 3


import ZODB # for "cPersistence"
from _isearch import ISearch, Unspecified, _IBTree, _IAnd, _IOr


##############################################################################
##############################################################################
## "BTrees" as incremental searches
##############################################################################
##############################################################################

from inspect import getmro

from BTrees.OOBTree import OOBTree, OOBucket, OOSet, OOTreeSet
from BTrees.IOBTree import IOBTree, IOBucket, IOSet, IOTreeSet
from BTrees.OIBTree import OIBTree, OIBucket, OISet, OITreeSet
from BTrees.IIBTree import IIBTree, IIBucket, IISet, IITreeSet
from BTrees.IFBTree import IFBTree, IFBucket, IFSet, IFTreeSet
from BTrees.LOBTree import LOBTree, LOBucket, LOSet, LOTreeSet
from BTrees.OLBTree import OLBTree, OLBucket, OLSet, OLTreeSet
from BTrees.LLBTree import LLBTree, LLBucket, LLSet, LLTreeSet
from BTrees.LFBTree import LFBTree, LFBucket, LFSet, LFTreeSet


# classification: type -> (maxLeafSize, maxNodeSize, keyType)
# size values from C code

_btree_classifier = {
  type(OOBTree()) : (30, 250, OBJECT,),
  type(IOBTree()) : (60, 500, INT,),
  type(OIBTree()) : (60, 250, OBJECT,),
  type(IIBTree()) : (120, 500, INT,),
  type(IFBTree()) : (120, 500, INT,),
  type(LOBTree()) : (60, 500, LONG,),
  type(OLBTree()) : (60, 250, OBJECT,),
  type(LLBTree()) : (120, 500, LONG,),
  type(LFBTree()) : (120, 500, LONG,),
  type(OOBucket()) : (0, 0, OBJECT,),
  type(IOBucket()) : (0, 0, INT,),
  type(OIBucket()) : (0, 0, OBJECT,),
  type(IIBucket()) : (0, 0, INT,),
  type(IFBucket()) : (0, 0, INT,),
  type(LOBucket()) : (0, 0, LONG,),
  type(OLBucket()) : (0, 0, OBJECT,),
  type(LLBucket()) : (0, 0, LONG,),
  type(LFBucket()) : (0, 0, LONG,),
  type(OOTreeSet()) : (30, 250, OBJECT,),
  type(IITreeSet()) : (30, 250, INT,),
  type(LLTreeSet()) : (30, 250, LONG,),
  type(OOSet()) : (0, 0, OBJECT,),
  type(IISet()) : (0, 0, INT,),
  type(LLSet()) : (0, 0, LONG,),
  }

def classifyBTree(tree):
  '''classify *tree*.

  raises a 'TypeError' when *tree* does not derive from
  a known 'BTree' type.
  '''
  cl = _btree_classifier
  for t in getmro(type(tree)):
    if t in cl: return cl[t]
  raise TypeError("not derived from a supported 'BTrees' type")


class IBTree(_IBTree):
  '''some BTrees instance as an incremental search.'''
  def __init__(self, btree, keytype=None):
    # raise exception, if type unknown
    maxLeafSize, maxNodeSize, keyType = classifyBTree(btree)
    if keytype is None: keytype = keyType
    elif keytype != keyType:
      raise TypeError('Incompatible key types')
    _IBTree.__init__(self, keytype, btree,
                     3 * maxNodeSize // 4,
                     3 * maxLeafSize // 4,
                     )


##############################################################################
##############################################################################
## "IAnd" and "IOr"
##############################################################################
##############################################################################


class _MultiSearch(object):
  '''abstract base class for 'IAnd' and 'IOr'.

  A combined search over one or more ISearches.
  '''

  _searches = None

  def __init__(self, keytype, *searches):
    self._keytype = keytype
    self._searches = []
    for s in searches: self.addSearch(s)
  
  def addSearch(self, search):
    if self._keytype != search.keytype:
      raise ValueError('disagreeing keytypes')
    self._searches.append(search)

  def complete(self):
    raise NotImplementedError


class _IEmpty(ISearch):
  '''the empty ISearch.'''
  def __init__(self, keytype):
    ISearch.__init__(self, keytype)
    self.classification = AT_END
    self.estimatedSize = 0

  def advanceTo(self, to): return AT_END
  def advanceFrom(self, start, limit): return AT_END


class IAnd(_MultiSearch, _IAnd):
  '''the incremental and of one or more ISearches.'''
  def complete(self):
    searches = self._searches
    if searches is None:
      raise ValueError("must not call 'complete' twice")
    # eliminate nested ands and ors with less than 2 elements.
    i = 0
    while i < len(searches):
      search = searches[i]
      if isinstance(search, _IEmpty) \
         or isinstance(search, IOr) and not search.size:
        searches[:] = [_IEmpty(self._keytype)]; break
      if isinstance(search, _IOr) and search.size == 1:
        searches[i] = search.unmantle()[0]; continue
      if isinstance(search, _IAnd):
        searches[i:i+1] = search.unmantle(); continue
      i += 1
    # order by increasing size -- indeterminate is considered very large
    # also estimate size as the minimal size
    from sys import maxint
    l = []; estimatedSize = maxint
    for s in searches:
      size = s.estimatedSize
      if size < 0: size = maxint
      elif size < estimatedSize: estimatedSize = size
      l.append((size, s))
    l.sort()
    if estimatedSize == maxint: estimatedSize = -1 # no useful estimation
    del self._searches # prevent a second use
    _IAnd.__init__(self, self._keytype, [ss[1] for ss in l])
    self.estimatedSize = estimatedSize
    self.size = len(l)

class IAnd_int(IAnd):
  ''''IAnd' specialization for 'INT' keys.'''
  def __init__(self, *searches):
    IAnd.__init__(self, INT, *searches)

class IAnd_obj(IAnd):
  ''''IAnd' specialization for 'OBJECT' keys.'''
  def __init__(self, *searches):
    IAnd.__init__(self, OBJECT, *searches)

class IAnd_Long(IAnd):
  ''''IAnd' specialization for 'LONG' keys.'''
  def __init__(self, *searches):
    IAnd.__init__(self, LONG, *searches)

class IOr(_MultiSearch, _IOr):
  '''the incremental or of zero or more ISearches.'''
  def complete(self):
    searches = self._searches
    # eliminate nested ors and one element ands
    i = 0
    while i < len(searches):
      search = searches[i]
      if isinstance(search, _IEmpty): del searches[i]; continue
      if isinstance(search, _IAnd) and search.size == 1:
        searches[i] = search.unmantle()[0]; continue
      if isinstance(search, _IOr):
        searches[i:i+1] = search.unmantle(); continue
      i += 1
    # order be descreasing size -- indeterminate is considered very small
    # estimate the size as the sum of the determinate sizes
    l = []; estimatedSize = 0; indeterminate = True
    for s in searches:
      size = s.estimatedSize
      if size >= 0: indeterminate = False; estimatedSize += size
      l.append((size, s))
    l.sort(); l.reverse()
    del self._searches # prevent a second use
    _IOr.__init__(self, self._keytype, [ss[1] for ss in l])
    self.estimatedSize = estimatedSize
    self.size = len(l)

class IOr_int(IOr):
  ''''IOr' specialization for 'INT' keys.'''
  def __init__(self, *searches):
    IOr.__init__(self, INT, *searches)

class IOr_obj(IOr):
  ''''IOr' specialization for 'OBJECT' keys.'''
  def __init__(self, *searches):
    IOr.__init__(self, OBJECT, *searches)


##############################################################################
##############################################################################
## INot
##############################################################################
##############################################################################


class INot(ISearch):
  _search = None

  def __init__(self, search, enumerator=None):
    ISearch.__init__(self, search.keytype)
    self._search = search
    self._enumerator = enumerator

  def advanceTo(self, to):
    '''we assume without check that *to* is in the domain.'''
    search = self._search; self.value = to
    if search is None: cl = INLIST_SUCCESS
    else:
      cl = search.advanceTo(to)
      if cl == AT_END: del self._search
      cl = cl == INLIST_SUCCESS and NOT_INLIST or INLIST_SUCCESS
    self.classification = cl
    return cl

  def advanceFrom(self, fromKey, limitKey):
    search = self._search; enumerator = self._enumerator
    next = fromKey
    while True:
      next = enumerator.next(next)
      if next is Unspecified: cl = AT_END; break
      if limitKey is not Unspecified and cmp(next, limitKey) >= 0:
        cl = CANDIDATE; break
      if search is None: cl = INLIST_SUCCESS; break
      cl = search.advanceTo(next)
      if cl == AT_END: del self._search
      if cl != INLIST_SUCCESS: cl = INLIST_SUCCESS; break
    if cl != AT_END: self.value = next
    self.classification = cl
    return cl


##############################################################################
##############################################################################
## IFilter
##############################################################################
##############################################################################


class IFilter(ISearch):

  def __init__(self, keytype, filter, enumerator=None):
    ISearch.__init__(self, keytype)
    self._filter = filter
    self._enumerator = enumerator

  def advanceTo(self, to):
    '''we assume without check that *to* is in the domain.'''
    filter = self._filter; self.value = to
    fr = filter(to)
    if fr: cl = INLIST_SUCCESS
    else: cl = NOT_INLIST
    self.classification = cl
    return cl

  def advanceFrom(self, fromKey, limitKey):
    filter = self._filter; enumerator = self._enumerator
    next = fromKey
    while True:
      next = enumerator.next(next)
      if next is Unspecified: cl = AT_END; break
      if limitKey is not Unspecified and cmp(next, limitKey) >= 0:
        cl = CANDIDATE; break
      fr = filter(next)
      if fr: cl = INLIST_SUCCESS; break
    if cl != AT_END: self.value = next
    self.classification = cl
    return cl

class IFilter_int(IFilter):
  ''''IFilter' specialization for 'INT' keys.'''
  def __init__(self, filter, enumerator=None):
    IFilter.__init__(self, INT, filter, enumerator)

class IFilter_obj(IFilter):
  ''''IFilter' specialization for 'OBJECT' keys.'''
  def __init__(self, filter, enumerator=None):
    IFilter.__init__(self, OBJECT, filter, enumerator)

class IFilter_long(IFilter):
  ''''IFilter' specialization for 'OBJECT' keys.'''
  def __init__(self, filter, enumerator=None):
    IFilter.__init__(self, LONG, filter, enumerator)


##############################################################################
##############################################################################
## Enumerator
##############################################################################
##############################################################################

class Enumerator(object):
  '''turn a 'BTrees' object into a document enumerator.

  Note, this fails if the tree contains 'None' as key (bug in
  the 'BTrees' specification.
  '''
  def __init__(self, btree):
    self._tree = btree

  def check(self, elem):
    return self._tree.has_key(elem)

  def next(self, elem):
    if elem is Unspecified: elem = None
    candidates = self._tree.keys(elem)
    # from ZODB 3.3 on, we can specify that we do not want "elem"
    if not candidates: return Unspecified
    candidate = candidates[0]
    if candidate != elem: return candidate
    try: return candidates[1]
    except IndexError: return Unspecified
  



##############################################################################
##############################################################################
## intersection
##############################################################################
##############################################################################

def intersection(keytype, *sets):
  '''the intersection of *sets*.

  *sets* must not be empty; any set must agree with *keytype*.
  '''
  iand = IAnd(keytype, *[IBTree(set, keytype) for set in sets])
  iand.complete()
  return iand.asSet()

def intersection_int(*sets):
  return intersection(INT, *sets)

def intersection_obj(*sets):
  return intersection(OBJECT, *sets)

def intersection_long(*sets):
  return intersection(LONG, *sets)




#       $Id: setup.py,v 1.1.1.1 2008/06/27 19:21:39 dieter Exp $
from os import environ
from os.path import join, exists
from sys import exit
from distutils.core import Extension

ZODB_HOME = environ.get('ZODB_HOME')
if not ZODB_HOME:
  exit("""This package depends on the ZODB3 source. 
    You must make this source available and then let the environment variable
    'ZODB_HOME' point to it.
    If you use the ZODB source within a Zope2 source tree, let 'ZODB_HOME'
    point to the 'lib/python' subdirectory of your Zope2 source tree.
    """)
# normalize a bit
#  the 'easy_install -e' download contains the source code inside an 'src'
#  subdirectory. ZODB sources which are part of "Zope2" lack this 'src'.
if exists(join(ZODB_HOME, 'src')): ZODB_HOME = join(ZODB_HOME, 'src')

if not exists(join(ZODB_HOME, "persistent", "cPersistence.h")):
  exit("cannot recognize the ZODB source")

from os.path import abspath, dirname, join
try:
  # try to use setuptools
  from setuptools import setup, Extension
  setupArgs = dict(
      install_requires=[], # we need the ZODB3 source but do not know how to express this
      include_package_data=True,
      namespace_packages=['dm'],
      zip_safe=False, # to let the tests work
      )
except ImportError:
  # use distutils
  from distutils import setup
  from distutils.core import Extension
  setupArgs = dict(
    )

cd = abspath(dirname(__file__))
pd = join(cd, 'dm', 'incrementalsearch')

def pread(filename, base=pd): return open(join(base, filename)).read().rstrip()


setup(name = "dm.incrementalsearch",
      version=pread('VERSION.txt').split('\n')[0],
      description='An efficient low level search execution engine on top of ZODB.BTrees.',
      long_description=pread('README.txt'),
      classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Framework :: ZODB',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: C',
        'Topic :: Utilities',
        ],
      author='Dieter Maurer',
      author_email='dieter@handshake.de',
      url='http://www.dieter.handshake.de/pyprojects/zope',
      packages=['dm', 'dm.incrementalsearch', 'dm.incrementalsearch.tests'],
      keywords='search, efficient, ZODB, incremental, BTrees',
      license='BSD (see "dm/incrementalsearch/LICENSE.txt", for details)',
      ext_modules = [
         Extension("dm.incrementalsearch._isearch",
            [join('dm', 'incrementalsearch', f)
             for f in ("_isearch.c", "_isearch_int.c", "_isearch_obj.c", "_isearch_long.c")],
            include_dirs = [ZODB_HOME,],
            ),
         ],
      **setupArgs
      )
